"""
    Copyright (C) <2021>  <ADP, F. Hartmann>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import json
import os
import sys
import time

import wx

THISFILE = os.path.abspath(sys.argv[0])
THISPATH = os.path.dirname(THISFILE)
if THISPATH not in sys.path:
    sys.path.append(THISPATH)

if sys.version_info.major > 2:
    # Python 3 (Poser 12 and above)
    basestring = str
    map = lambda a, b: [a(_b) for _b in b]
    from importlib import reload
else:
    # Python 2 (Poser 11 and below)
    if not os.path.exists(os.path.join(THISPATH, "__init__.py")):
        # this file must exist to be able to make includes
        with open(os.path.join(THISPATH, "__init__.py"), "w"):
            pass

try:
    import poser
except ImportError:
    import POSER_FAKE as poser

from local_libs import im_export, WXtools
from local_libs.im_export import do_import, do_export

reload(WXtools)
aui = poser.WxAuiManager()

SCENE = poser.Scene()
HDRI_SUPPORT = False
TITLE = "OBJ Im-/Export"
TEMPDIR = poser.TempLocation()
IMEX_CONFIG = globals().setdefault("IMEX_CONFIG", dict()) or dict()
IMEX_CONFIG["THISPATH"] = os.path.dirname(sys.argv[0])
im_export.IMEX_CONFIG = IMEX_CONFIG

LIGHT = 0xc0, 0xc0, 0xc0
DARK = 0x4b, 0x4b, 0x4b
BTN_BG_COLOR = 0x40, 0x40, 0x40
CHOICE_BG_COLOR = WXtools.changeLightness(DARK, 110)
INPUT_BG_COLOR = 0x60, 0x60, 0x60
INPUT_FG_COLOR = 0xc0, 0xee, 0xff


def load_config():
    global IMEX_CONFIG
    fname = os.path.join(THISPATH, "imex.config")
    if os.path.exists(fname) and os.access(fname, os.R_OK):
        with open(os.path.join(THISPATH, "imex.config")) as fh:
            IMEX_CONFIG = json.load(fh)
        return True
    return False


def save_config():
    global IMEX_CONFIG
    fname = os.path.join(THISPATH, "imex.config")
    with open(fname, "w") as fh:
        json.dump(IMEX_CONFIG, fh, indent=4)


class Menu(wx.Panel):
    _CONFIGCHANGED = False

    def __init__(self, parent,
                 name=TITLE,
                 style=wx.DEFAULT_FRAME_STYLE,
                 size=(270, 300),
                 **kwargs):

        WXtools.CONFIG = IMEX_CONFIG

        super(self.__class__, self).__init__(parent=parent,
                                             id=wx.ID_ANY,
                                             name=name,
                                             style=style,
                                             size=size,
                                             **kwargs)
        load_config()
        IMEX_CONFIG["BackgroundColor"] = DARK
        IMEX_CONFIG["ForegroundColor"] = LIGHT
        self.SetBackgroundColour(DARK)
        self.SetForegroundColour(LIGHT)

        self.gb_manager = WXtools.GridBagManager(parent=self,
                                                 sizer=wx.GridBagSizer(2, 4),
                                                 flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL | wx.EXPAND,
                                                 border=2
                                                 )

        self.figurelist = WXtools.PrepCtrl(WXtools.ChoiceCtrl, choices=[],
                                           name="figurelist", bgcolor=CHOICE_BG_COLOR,
                                           tooltip="Left click to select.\n"
                                                   "Right click to reload Figurelist.",
                                           bind=((wx.EVT_RIGHT_UP, self.updateFigurelist),
                                                 (WXtools.EVT_POPUP_SELECTION, self.onChoice))
                                           )
        self.SetAutoLayout(True)
        self.initUI(self.gb_manager)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.onClose)
        self.Bind(WXtools.EVT_CHECK_CHANGED, self.onCheckbox)
        self.Bind(wx.EVT_CHECKBOX, self.onCheckbox)
        self.updateFigurelist()

    def initUI(self, manager):
        isinstance(manager, WXtools.GridBagManager)
        labelfont = root.GetFont()  # type: wx.Font
        labelfont.SetWeight(wx.FONTWEIGHT_BOLD)
        add = manager.addrow
        add(WXtools.LabelCtrl(label="Auto Sync", font=labelfont),
            WXtools.PrepCtrl(WXtools.CheckboxCtrl, name="FLAG_autoSync",
                             tooltip="Sync Figure with Poser Selection",
                             style=wx.ALIGN_RIGHT | wx.NO_BORDER,
                             bgcolor=DARK,
                             value=IMEX_CONFIG.setdefault("FLAG_autoSync", False)),
            flag=wx.EXPAND)
        add(WXtools.LabelCtrl(label="Figure", font=labelfont), self.figurelist)
        add(None)
        add(WXtools.LabelCtrl(label="Scale", font=labelfont),
            WXtools.PrepCtrl(wx.TextCtrl, name="TX_Scale",
                             bgcolor=INPUT_BG_COLOR, fgcolor=INPUT_FG_COLOR,
                             tooltip="Scale for OBJ export and import.\n"
                                     "Insert a number in the range 1-1000",
                             style=wx.ALIGN_RIGHT | wx.BORDER_NONE,
                             value=str(IMEX_CONFIG.get("TX_Scale", 1))),
            flag=wx.EXPAND)
        add(WXtools.LabelCtrl(label="Export Groups", font=labelfont),
            WXtools.PrepCtrl(WXtools.CheckboxCtrl, name="FLAG_UseGroups",
                             tooltip="Write actor Groupnames into export file.",
                             style=wx.ALIGN_RIGHT | wx.BORDER_NONE,
                             value=IMEX_CONFIG.get("FLAG_UseGroups", False)))
        add(WXtools.LabelCtrl(label="Export UV info", font=labelfont),
            WXtools.PrepCtrl(WXtools.CheckboxCtrl, name="FLAG_UseMat",
                             tooltip="Write UV and Texture info to export file.",
                             style=wx.ALIGN_RIGHT | wx.BORDER_NONE,
                             value=IMEX_CONFIG.get("FLAG_UseMat", True)))
        add(WXtools.LabelCtrl(label="Copy Imagefiles", font=labelfont),
            WXtools.PrepCtrl(WXtools.CheckboxCtrl, name="FLAG_ImgCopy",
                             tooltip="Copy texture files to export path.",
                             style=wx.ALIGN_RIGHT,
                             value=IMEX_CONFIG.get("FLAG_ImgCopy", True)))
        add(WXtools.LabelCtrl(label="Auto Increment", font=labelfont),
            WXtools.PrepCtrl(WXtools.CheckboxCtrl, name="FLAG_autoIncrement",
                             tooltip="Increment morphname on each export.\n"
                                     "Adds an incrementing number to the name.",
                             style=wx.ALIGN_RIGHT | wx.NO_BORDER,
                             value=IMEX_CONFIG.setdefault("FLAG_autoIncrement", False)))

        add(None)
        add(WXtools.LabelCtrl(label="Morphname", font=labelfont),
            WXtools.PrepCtrl(wx.TextCtrl, name="TX_Morphname",
                             bgcolor=INPUT_BG_COLOR, fgcolor=INPUT_FG_COLOR,
                             tooltip="Morphname used in all actors for next imported morph",
                             style=wx.ALIGN_RIGHT | wx.BORDER_NONE,
                             value=self.get_Morphname()),
            flag=wx.EXPAND)
        add(WXtools.LabelCtrl(label="X Center == 0"),
            WXtools.PrepCtrl(WXtools.CheckboxCtrl, name="FLAG_ZeroCenter",
                             tooltip="Force Centerpoints on X-axis to be Zero while importing.",
                             style=wx.ALIGN_RIGHT,
                             value=IMEX_CONFIG.get("FLAG_ZeroCenter", False)))
        add(None)
        add(WXtools.LabelCtrl(label="Silent Im-/Export"),
            WXtools.PrepCtrl(WXtools.CheckboxCtrl, name="FLAG_SilentImExport",
                             tooltip="Im-/Export to/from file without Dialog if possible.",
                             style=wx.ALIGN_RIGHT,
                             value=IMEX_CONFIG.get("FLAG_SilentImExport", False)))
        add(None)
        add(WXtools.PrepCtrl(wx.Button, name="BTN_ExportFigure",
                             label="Export Figure",
                             tooltip="Export Figure as OBJ File\nto be morphed in external modeller.",
                             font=labelfont,
                             style=wx.BORDER_NONE,
                             bgcolor=BTN_BG_COLOR, fgcolor=(0xf0, 0xd0, 0x50),
                             bind=((wx.EVT_LEFT_UP, self.onExportModel),
                                   (wx.EVT_ENTER_WINDOW, self.onTooltip))
                             ),
            WXtools.PrepCtrl(wx.Button, name="BTN_ImportFigure",
                             label="Import Morph",
                             tooltip="Import OBJ File as Morph",
                             font=labelfont,
                             style=wx.BORDER_NONE,
                             bgcolor=BTN_BG_COLOR, fgcolor=(0xff, 0xa0, 0x30),
                             bind=((wx.EVT_LEFT_UP, self.onImportModel),
                                   (wx.EVT_ENTER_WINDOW, self.onTooltip))
                             )
            )

        manager.sizer.AddGrowableCol(1)
        sz = wx.BoxSizer(wx.VERTICAL)
        sz.Add(manager.sizer, 1, wx.EXPAND | wx.ALIGN_CENTER | wx.ALL, 4)

        if HDRI_SUPPORT:
            sz.Add(wx.StaticLine(self, wx.ID_ANY, size=(1, 1), style=wx.LI_HORIZONTAL),
                   0, wx.ALL | wx.EXPAND, 16)
            sz.Add(WXtools.PrepCtrl(wx.Button, name="BTN_Render",
                                    label="Lights off and Render!",
                                    tooltip="Switch all lights off and Render current image.",
                                    font=labelfont,
                                    style=wx.BORDER_NONE,
                                    bgcolor=BTN_BG_COLOR, fgcolor=(160, 160, 160),
                                    bind=(wx.EVT_LEFT_UP, self.onRender)),
                   0, wx.EXPAND | wx.ALIGN_CENTER_HORIZONTAL, 8)
        sz.AddStretchSpacer()
        self.SetSizerAndFit(sz)

    def todays_morphname(self):
        return "%s_%s" % (self.figurelist.GetStringSelection() or "MORPH", time.strftime("%m%d"))

    def updateFigurelist(self, ev=None):
        with wx.BusyInfo("Updating Figurelist"):
            self.get_current_config_collection()["TX_Morphname"] = self.get_Morphname()
            while self.figurelist.GetCount():
                self.figurelist.Delete(0)

            for idx, fig in enumerate(poser.Scene().Figures()):
                self.figurelist.Append(fig.Name())
                if fig == poser.Scene().CurrentFigure():
                    self.figurelist.SetSelection(idx)

            mname = self.get_current_config_collection().get("TX_Morphname", self.todays_morphname())
            self.FindWindowByName("TX_Morphname").SetValue(mname)

        self._CONFIGCHANGED = True

    def getScale(self):
        scale = wx.FindWindowByName("TX_Scale").GetValue()  # type: str
        return 1 if not scale.isnumeric() else int(scale)

    def get_current_config_collection(self):
        figurename = self.figurelist.GetStringSelection()
        return IMEX_CONFIG.get(figurename, dict()) if figurename else dict()

    def get_Morphname(self):
        widget = wx.FindWindowByName("TX_Morphname")
        if widget:
            return widget.GetValue()
        return self.get_current_config_collection().get("TX_Morphname", self.todays_morphname())

    def incremented_morphname(self, mname=None):
        if IMEX_CONFIG.get("FLAG_autoIncrement", False):
            if mname is None:
                mname = wx.FindWindowByName("TX_Morphname").GetValue()
            if "." in mname:
                mname, _, nr = mname.rpartition(".")
            else:
                nr = "0"
            try:
                nr = int(nr)
            except Exception:
                mname = wx.FindWindowByName("TX_Morphname").GetValue()
                nr = 0
            return "%s.%02d" % (mname, nr + 1)
        return None

    def onChoice(self, ev=None):
        selection = ev.selection
        if IMEX_CONFIG.get("FLAG_autoSync", False):
            try:
                fig = SCENE.Figure(selection)
            except poser.error:
                return
            else:
                SCENE.SelectFigure(fig)
        self.figurelist.SetSelection(ev.idx)
        mname = self.get_current_config_collection().get("TX_Morphname", self.todays_morphname())
        wx.FindWindowByName("TX_Morphname").SetValue(mname)


    def onCheckbox(self, ev):
        obj = ev.GetEventObject()
        IMEX_CONFIG[obj.Name] = ev.checked
        self._CONFIGCHANGED = True
        ev.Skip()

    def onClose(self, ev=None):
        try:
            IMEX_CONFIG["Perspective"] = aui.SavePaneInfo(paneInfo)
        except Exception as err:
            pass
        # force final saving
        self._CONFIGCHANGED = False
        IMEX_CONFIG.update(self.gb_manager.items_as_dict())
        save_config()

    def onTooltip(self, event):
        obj = event.GetEventObject()
        config_collection = self.get_current_config_collection()
        tx = config_collection.get("out_fname", "") if obj.Name == "BTN_ExportFigure" \
            else config_collection.get("in_fname", "") if obj.Name == "BTN_ImportFigure" \
            else None
        if tx:
            tt = obj.GetToolTip()
            s = tt.GetTip()
            tt.SetTip(s[:s.rfind("\npath:")] + "\npath:" + tx)
        event.Skip()

    def onExportModel(self, ev=None):
        figurename = self.figurelist.GetStringSelection()
        if figurename is not None:
            aborted = False
            silent = wx.FindWindowByName("FLAG_SilentImExport", self).IsChecked()

            config_collection = self.get_current_config_collection()  # filenames and parameters for one figure
            out_fname = config_collection.get("out_fname", figurename + "_Export.obj")
            if not silent or not out_fname or not os.path.isdir(config_collection.get("last_path", "")):
                with wx.FileDialog(self, message="Select file for Export",
                                   defaultDir=config_collection.get("last_path", ""),
                                   defaultFile=out_fname,
                                   wildcard="OBJ files (*.obj)|*.obj",
                                   style=wx.FD_DEFAULT_STYLE | wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT
                                   ) as dlg:
                    if dlg.ShowModal() == wx.ID_CANCEL:
                        aborted = True
                    else:
                        out_fname = dlg.GetPath()
                        config_collection["last_path"] = os.path.dirname(out_fname)
                        config_collection["out_fname"] = out_fname

            if not aborted:
                SCENE.DrawAll()
                with wx.BusyInfo("Exporting '%s' to %s" % (figurename, config_collection.get("out_fname"))):
                    figure = SCENE.Figure(figurename)
                    IMEX_CONFIG["CTRL_MoveTextures"] = wx.FindWindowByName("FLAG_ImgCopy").IsChecked()
                    groups = IMEX_CONFIG["FLAG_UseGroups"] = wx.FindWindowByName("FLAG_UseGroups").IsChecked()
                    mats = IMEX_CONFIG["FLAG_UseMat"] = wx.FindWindowByName("FLAG_UseMat").IsChecked()
                    res = do_export(figure,
                                    filename=out_fname,
                                    scale=self.getScale(),
                                    use_groups=groups,
                                    use_mat=mats)

            IMEX_CONFIG[figurename] = config_collection
            mname = self.incremented_morphname()
            if mname:
                self.FindWindowByName("TX_Morphname").SetValue(mname)
                self.get_current_config_collection()["TX_Morphname"] = mname
        else:
            WXtools.PopUpMessage("Figure '%s' not in Scene." % figurename,
                                 closeafter=10,
                                 fg=(255, 255, 255), bg=(0xf0, 40, 40))

        self._CONFIGCHANGED = True
        ev.Skip()

    def onImportModel(self, ev=None):
        figurename = self.figurelist.GetStringSelection()
        if figurename is None:
            return WXtools.PopUpMessage("Figure '%s' not in Scene." % figurename,
                                        closeafter=10,
                                        fg=(255, 255, 255), bg=(0xf0, 40, 40))

        config_collection = IMEX_CONFIG.get(figurename, dict())
        try:
            SCENE.SelectFigure(SCENE.Figure(figurename))
        except poser.error:
            self.updateFigurelist(None)
            return WXtools.PopUpMessage("Figure '%s' not in Scene." % figurename,
                                        closeafter=10,
                                        fg=(255, 255, 255), bg=(0xf0, 40, 40))

        aborted = False
        silent = wx.FindWindowByName("FLAG_SilentImExport", self).IsChecked()

        in_fname = config_collection.get("in_fname", "")
        if not silent or not in_fname or not os.path.isdir(config_collection.get("last_path", "")):
            with wx.FileDialog(self, message="Select file for Import",
                               defaultDir=config_collection.get("last_path", THISPATH),
                               defaultFile=in_fname,
                               wildcard="OBJ files (*.obj)|*.obj",
                               style=wx.FD_DEFAULT_STYLE | wx.FD_OPEN
                               ) as dlg:
                if dlg.ShowModal() == wx.ID_CANCEL:
                    aborted = True
                else:
                    in_fname = dlg.GetPath()
                    config_collection["last_path"] = os.path.dirname(in_fname)
                    config_collection["in_fname"] = in_fname

        if not aborted:
            res = None
            if os.path.isfile(in_fname):
                with wx.BusyInfo("Importing Morph to Figure '%s'" % figurename):
                    try:
                        res = do_import(SCENE.CurrentFigure(),
                                        new_filename=in_fname,
                                        old_filename=config_collection.get("out_fname"),
                                        morphname=wx.FindWindowByName("TX_Morphname").GetValue(),
                                        force_center=wx.FindWindowByName("FLAG_ZeroCenter", self).IsChecked(),
                                        scale=self.getScale(),
                                        )
                    except Exception:
                        return

            IMEX_CONFIG[figurename] = config_collection
            if res is None:
                WXtools.PopUpMessage("OBJ file not found or wrong file used.", closeafter=5)

        self._CONFIGCHANGED = True
        ev.Skip()

    def onRender(self, ev):
        self.lights = []
        for ac in SCENE.Lights():  # type: poser.ActorType
            self.lights.append((ac.InternalName(), ac.LightOn()))
            ac.SetLightOn(False)
        SCENE.Render()
        for ac, flag in self.lights:
            SCENE.ActorByInternalName(ac).SetLightOn(flag)


__timer__ = wx.Timer()


def tt():
    global __timer__
    try:
        if MainPanel and IMEX_CONFIG.get("FLAG_autoSync"):
            cf = SCENE.CurrentFigure()
            if cf:
                figurename = MainPanel.figurelist.GetStringSelection()
                if figurename != cf.Name():
                    MainPanel.updateFigurelist()
    except Exception:
        pass
    try:
        if MainPanel._CONFIGCHANGED:
            MainPanel._CONFIGCHANGED = False
        save_config()
        __timer__ = wx.CallLater(1000, tt)

    except Exception as err:
        __timer__.Stop()


if __name__ == "__main__":
    import wx.aui

    aui = poser.WxAuiManager()
    root = aui.GetManagedWindow()

    if not load_config():
        IMEX_CONFIG.update(dict(
            MORPH_ADD_TO_OLD_VALUE=True,
            Scale=1,
        ))

    paneInfo = aui.GetPane(TITLE)  # type: wx.aui.AuiPaneInfo
    if paneInfo.IsOk():
        aui.ClosePane(paneInfo)
        paneInfo = aui.GetPane(TITLE)

    savedinfo = IMEX_CONFIG.get("Perspective", None)
    if savedinfo is not None:
        aui.LoadPaneInfo(savedinfo, paneInfo)
    else:
        paneInfo.name = TITLE
        paneInfo.Movable(True).DestroyOnClose(True)

    paneInfo.Name(TITLE).Caption(TITLE)

    MainPanel = Menu(parent=root, name=TITLE)
    aui.AddPane(MainPanel, paneInfo)
    paneInfo.Show()
    tt()
    aui.Update()
