"""
    Copyright (C) <2021>  <ADP, F. Hartmann>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import print_function

import wx
from wx.lib import buttons
from wxPrepare import *

LABEL_ARGS = dict()  # Arguments used for wx.StaticText
CTRL_ARGS = dict()  # Arguments used for wx.<Control>

try:
    import poser
except ImportError:
    import POSER_FAKE as poser

aui = poser.WxAuiManager()
root = aui.GetManagedWindow()
SYS_FONT = root.GetFont()
CURRENT_GRB_MGR = None
GLOBAL_PARENTWINDOW = None
CONFIG = dict(BackgroundColor=wx.SYS_COLOUR_BACKGROUND,
              ForegroundColor=color_gray(0x40))

PopupEvent, EVT_POPUP_SELECTION = newevent.NewEvent()
CheckBoxEvent, EVT_CHECK_CHANGED = newevent.NewEvent()


class GridBagManager(object):
    __slots__ = "parent", "widgets", "sizer", "flag", "border", "current_row", "font"

    def __init__(self, *args, **kwargs):
        global CURRENT_GRB_MGR, GLOBAL_PARENTWINDOW

        for varname in self.__slots__:
            setattr(self, varname, kwargs.get(varname, None))

        for idx, entry in enumerate(args):
            if self.__slots__[idx] not in kwargs:
                setattr(self, self.__slots__[idx], entry)

        self.current_row = 0
        self.widgets = []
        CURRENT_GRB_MGR = self
        GLOBAL_PARENTWINDOW = self.parent

    def addrow(self, *args, **kwargs):
        row = int(kwargs.get("row", self.current_row))
        col = int(kwargs.get("startcol", 0))
        flag = kwargs.get("flag", self.flag)
        border = kwargs.get("border", self.border)
        font = kwargs.pop("font", SYS_FONT)
        sz = self.sizer  # type: wx.GridBagSizer
        for idx, widget in enumerate(args):
            if widget is None:
                sz.AddGrowableRow(row)
            else:
                if font and not isinstance(widget, wx.StaticLine):
                    widget.SetFont(font)

                self.sizer.Add(widget,
                               pos=(row, col + idx),
                               flag=flag,
                               border=border)
                if widget.__class__ != TextLabel:
                    self.widgets.append(widget)

        self.current_row += 1

    def get_items_by_name(self, name, filter=None):
        if not filter:
            return [w for w in self.widgets if w.GetName().startswith(name)]
        return [w for w in self.widgets if w.GetName().startswith(name)
                and w.__class__.__name__ in filter]

    def get_value_by_name(self, name, filter=None):
        obj = self.get_items_by_name(name, filter)
        if not obj:
            return None
        if len(obj) == 1:
            return obj[0].Value()
        return [o.Value() for o in obj]

    def items_as_dict(self, filter=None):
        d = dict()
        for widget in [w for w in self.widgets
                       if filter is None or w.__class__.__name__ in filter]:
            if hasattr(widget, "GetCheckedStrings"):
                d[widget.GetName()] = widget.GetCheckedStrings()
            elif hasattr(widget, "GetValue"):
                d[widget.GetName()] = widget.GetValue()
            elif hasattr(widget, "GetSelection"):
                v = widget.GetSelection()
                if hasattr(widget, "GetString"):
                    v = widget.GetString(v)
                d[widget.GetName()] = v

        return d

    def addrows(self, *rows2add):
        for row in rows2add:
            if row is None:
                self.addrow(None)
            else:
                self.addrow(*row)

    def setStyles(self, ctrltype, **kwargs):
        for c in self.widgets:
            if c.__class__ is ctrltype:
                c.SetWindowStyle(**kwargs)

    def bindCtrls(self, ctrltype, ev, target):
        for c in self.widgets:
            if c.__class__ is ctrltype:
                c.Bind(ev, target)


def setTooltip(ctrl, text, expiretime=20000):
    if text and ctrl:
        t = wx.ToolTip(text)
        t.SetAutoPop(expiretime)
        ctrl.SetToolTip(t)
    return ctrl


def ColoredCtrl(ctrl, **kwargs):
    ctrl.SetBackgroundColour(kwargs.pop("bgcolor", CONFIG.get("BackgroundColor")))
    ctrl.SetForegroundColour(kwargs.pop("fgcolor", CONFIG.get("ForegroundColor")))
    return setTooltip(ctrl, kwargs.pop("tooltip", None))


def LabelCtrl(**kwargs):
    parent = GLOBAL_PARENTWINDOW
    bg_color = kwargs.pop("bgcolor", CONFIG.get("BackgroundColor"))
    fg_color = kwargs.pop("fgcolor", CONFIG.get("ForegroundColor"))
    font = kwargs.pop("font", None)
    tooltip = kwargs.pop("tooltip", None)
    ctrl = TextLabel(parent, wx.ID_ANY, **kwargs)
    if font:
        ctrl.SetFont(font)
    ctrl.SetBackgroundColour(bg_color)
    ctrl.SetForegroundColour(fg_color)
    return setTooltip(ctrl, tooltip)


def PrepCtrl(ctrl, **kwargs):
    parent = GLOBAL_PARENTWINDOW
    bind = kwargs.pop("bind", [])
    fgcolor = kwargs.pop("fgcolor", CONFIG.get("ForegroundColor"))
    bgcolor = kwargs.pop("bgcolor", CONFIG.get("BackgroundColor"))
    font = kwargs.pop("font", None)
    value = kwargs.pop("value", None)
    index = kwargs.pop("index", None)
    tooltip = kwargs.pop("tooltip", None)
    if "key" in kwargs:
        kwargs["name"] = kwargs["key"]
        kwargs.pop("key")
    ctrl = ctrl(parent, **kwargs)

    if bind:
        if isinstance(bind[0], (tuple, list)):
            for bind_entry in bind:
                assert len(bind_entry) == 2 and callable(bind_entry[1]), \
                    "Bind entries must be: int(EVT_ID), func<callable>"
                ctrl.Bind(bind_entry[0], bind_entry[1])
        else:
            assert len(bind) == 2 and callable(bind[1]), \
                "Bind entries must be: int(EVT_ID), func<callable>"
            ctrl.Bind(bind[0], bind[1])

    ctrl.SetForegroundColour(fgcolor)
    ctrl.SetBackgroundColour(bgcolor)
    if font:
        ctrl.SetFont(font)
    if value:
        ctrl.SetValue(value)
    if index:
        ctrl.SetSelection(index)
    return setTooltip(ctrl, tooltip)


def Labeled(ctrl, tooltip=None, **kwargs):
    label_args = LABEL_ARGS.copy()
    tx = kwargs.pop("text", kwargs.pop("label", ""))
    if tx:
        label_args["label"] = tx
    ctrl_args = CTRL_ARGS.copy()
    for k in kwargs.keys():
        if k.startswith("lbl_"):
            label_args[k[4:]] = kwargs.pop(k)
        elif k.startswith("ctl_"):
            ctrl_args[k[4:]] = kwargs.pop(k)
    ctrl_args.update(kwargs)

    return (setTooltip(LabelCtrl(**label_args), tooltip),
            setTooltip(PrepCtrl(ctrl=ctrl, **ctrl_args), tooltip))


def ButtonRow(*buttons, **sizerflags):
    parent = GLOBAL_PARENTWINDOW
    sz = wx.BoxSizer(wx.LI_HORIZONTAL)
    for btn in buttons:
        if isinstance(btn, basestring):
            btn = wx.Button(parent, -1, btn)
        sz.Add(btn, **sizerflags)

    sz.Layout()
    return sz


class TextLabel(wx.StaticText):
    def __init__(self, *args, **kwargs):
        wx.StaticText.__init__(self, *args, **kwargs)
        self.Bind(wx.EVT_PAINT, self.on_paint)
        tw, th = self.GetTextExtent(self.GetLabel())
        self.SetMinSize((tw + 10, th))

    def on_paint(self, event):
        self.do_draw(wx.BufferedPaintDC(self))

    def do_draw(self, dc):
        """:type dc: wx.PaintDC"""
        bgcolor = self.GetBackgroundColour()
        dc.SetBackground(wx.Brush(bgcolor, wx.BRUSHSTYLE_SOLID))
        dc.Clear()

        w, h = self.GetClientSize()
        dc.SetBrush(wx.Brush(changeLightness(bgcolor, 110), wx.BRUSHSTYLE_SOLID))
        dc.SetPen(wx.Pen(bgcolor, 0, wx.PENSTYLE_TRANSPARENT))
        dc.DrawRoundedRectangle(0, 0, w, h, -.5)
        dc.DrawRectangle(w - 20, 0, 20, h)

        tx = self.GetLabel()
        tw, th = dc.GetTextExtent(tx)
        top = (h - th) // 2
        dc.SetTextForeground(self.GetForegroundColour())
        dc.DrawText(tx, 8, top)


def PopUpMessage(text, bg=(0xf0, 0xf0, 0xf0), fg=(0x10, 0x10, 0x10), closeafter=3):
    root = wx.GetTopLevelWindows()[0]
    dialog = wx.Frame(root, id=wx.ID_ANY, title="IM-Export Notification",
                      style=wx.CAPTION | wx.CLOSE_BOX| wx.STAY_ON_TOP)
    dialog.SetForegroundColour(fg)
    dialog.SetBackgroundColour(bg)
    sz = wx.BoxSizer(wx.HORIZONTAL)
    txt = wx.StaticText(dialog, label=text)
    sz.Add(txt, 1, wx.ALL | wx.EXPAND, border=15)
    sz.Layout()
    dialog.SetSizer(sz)

    w, h = sz.ComputeFittingWindowSize(dialog)
    dialog.SetSize((w, h))

    pos_x, pos_y = root.GetPosition()
    size = root.GetSize()
    center_x = size[0] / 2.0
    center_y = size[1] / 2.0
    dialog.SetPosition((pos_x + center_x - w / 2, pos_y + center_y - h / 2))
    dialog.Show()

    if closeafter:
        wx.CallLater(closeafter * 1000, dialog.Destroy)

    return dialog


class PopMenu(wx.Menu):
    def __init__(self, parent, choices, **kwargs):
        """
        :type parent: wx.Window
        :type choices: list
        """
        super(PopMenu, self).__init__()
        self.parent = parent
        self.selected = 0
        self.choices = choices
        self.args = kwargs
        self.idlist = []
        for idx, entry in enumerate(choices):
            _id = newID()
            self.Append(_id, entry)
            self.idlist.append(_id)

        self.Bind(wx.EVT_MENU, self.on_sel)

    def on_sel(self, event):
        idx =  self.idlist.index(event.Id)
        ev = PopupEvent(idx=idx, selection=self.choices[idx])
        wx.PostEvent(self.parent, ev)


class ChoiceCtrl(buttons.GenButton):
    def __init__(self, parent, *args, **kwargs):
        self.choices = kwargs.pop("choices", [])
        super(ChoiceCtrl, self).__init__(parent, *args, style=kwargs.pop("style", 0))
        # All other provided parameters are handed over with the event from PopupMenu
        self.args = kwargs
        self.selected = None
        self.SetSelection(0)
        self.highlightcolor = (0xff, 0xa0, 0x30)
        self.fgcolor = (0xd0, 0xd0, 0xd0)
        self.bgcolor = kwargs.pop("bgcolor", parent.GetBackgroundColour())
        self.SetForegroundColour(self.fgcolor)
        self.SetBackgroundColour(self.bgcolor)
        self.up = True
        self.SetBezelWidth(0)
        self.Bind(wx.EVT_ENTER_WINDOW, self.on_enter)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.on_leave)
        self.Bind(wx.EVT_LEFT_UP, self.on_mouse)

    def on_enter(self, event):
        self.SetForegroundColour(self.highlightcolor)
        self.Refresh()

    def on_leave(self, event):
        self.SetForegroundColour(self.fgcolor)
        self.Refresh()

    def on_mouse(self, event):
        if self.choices:
            self.SetForegroundColour(self.fgcolor)
            popup = PopMenu(self, self.choices, **self.args)
            self.PopupMenu(popup, (0, self.GetClientSize()[1]))
            self.SetSelection(popup.selected)
            popup.Destroy()
        self.up = True
        self.Refresh()

    def DrawFocusIndicator(self, *args):
        return

    def DrawLabel(self, dc, width, height, dx=0, dy=0):
        super(ChoiceCtrl, self).DrawLabel(dc, width - 20, height, dx, dy)

        old_brush = dc.GetBrush()
        fgcolor = self.GetForegroundColour()
        bgcolor = self.GetBackgroundColour()

        brush = wx.Brush(changeLightness(bgcolor, 150), wx.BRUSHSTYLE_SOLID)
        pen = wx.Pen(fgcolor, 2, style=wx.PENSTYLE_TRANSPARENT)

        dc.SetPen(pen)
        dc.SetBrush(brush)
        w = 12
        h = 12

        dc.DrawPolygon([
            (0, 0),
            (w, 0),
            (w // 2, h)
        ], xoffset=width - 20, yoffset=h)

        dc.SetBrush(old_brush)

    def GetCount(self):
        return len(self.choices)

    def Delete(self, n):
        if self.choices:
            old = self.selected
            self.choices.pop(n)
            if n == old:
                self.SetSelection(0)

    def Append(self, s):
        self.choices.append(s)

    def SetSelection(self, n):
        lbl = "<List empty>"
        if self.choices:
            self.selected = n
            lbl = self.choices[n]
        self.SetLabel(lbl)
        self.Refresh()
        self.Update()

    def GetSelection(self):
        return self.selected

    def GetStringSelection(self):
        if self.choices:
            return self.choices[self.selected]
        return None


class CheckboxCtrl(buttons.GenButton):
    def __init__(self, parent, *args, **kwargs):
        """:type parent: wx.Window"""
        self.highlightcolor = kwargs.pop("hlcolor", (0xf0, 0xf0, 0xf0))
        self.fgcolor = kwargs.pop("fgcolor", parent.GetForegroundColour())
        self.bgcolor = kwargs.pop("bgcolor", parent.GetBackgroundColour())
        super(CheckboxCtrl, self).__init__(parent, *args, **kwargs)

        self.selected = False
        self.up = True
        self.SetForegroundColour(self.fgcolor)
        self.SetBackgroundColour(self.bgcolor)
        self.SetBezelWidth(0)
        self.Bind(wx.EVT_ENTER_WINDOW, self.on_enter)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.on_leave)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_mouse)

    def on_enter(self, event):
        self.SetForegroundColour(self.highlightcolor)
        self.Refresh()

    def on_leave(self, event):
        self.SetForegroundColour(self.fgcolor)
        self.Refresh()

    def on_mouse(self, event):
        self.up = True
        self.SetSelected(not self.selected)

    def IsSelected(self):
        return self.selected

    IsChecked = IsSelected

    def SetSelected(self, onoff):
        self.selected = bool(onoff)
        self.Refresh()
        self.Update()
        evt = CheckBoxEvent(GetEventObject=lambda: self, checked=self.selected)
        wx.PostEvent(self.GetParent(), evt)

    SetValue = SetSelected
    SetChecked = SetSelected

    def DrawFocusIndicator(self, *args):
        return

    def DrawLabel(self, dc, width, height, dx=0, dy=0):
        """:type dc: wx.PaintDC"""

        ofs = 4
        w = 16

        if self.GetLabel() != "":
            left = 0
            super(CheckboxCtrl, self).DrawLabel(dc, width - 20, height, dx, dy)
        else:
            left = width - (w + ofs)

        bgcolor = self.GetBackgroundColour()
        fgcolor = self.GetForegroundColour()

        brush = wx.Brush(changeLightness(bgcolor, 150), wx.BRUSHSTYLE_SOLID)
        pen = wx.Pen(fgcolor, 0, wx.PENSTYLE_TRANSPARENT)
        dc.SetBrush(brush)
        dc.SetPen(pen)

        dc.DrawRectangle(left, ofs, w, w)

        color = changeLightness(bgcolor, 150) if self.selected else bgcolor
        brush.SetColour(color)
        dc.SetBrush(brush)
        s = 2
        if self.selected:
            s = 3
            dc.SetPen(wx.Pen((0xe0, 0xa0, 0x30), 1, wx.PENSTYLE_SOLID))

        dc.DrawRectangle(left + s, ofs + s, w - s * 2, w - s * 2)
