"""
    Copyright (C) <2021>  <ADP, F. Hartmann>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import print_function

import sys
import wx
from wx.lib import newevent

if sys.version_info.major > 2:
    # Python 3 (Poser 12 and above)
    map = lambda a, b: [a(_b) for _b in b]
    basestring = str


    def wxCursor(*args):
        return wx.Cursor(*args)


    def newID():
        return int(wx.NewIdRef())


    def changeLightness(c, n):
        if c is not wx.Colour:
            c = wx.Colour(c)
        return c.ChangeLightness(n)


    def wxColour(*args, **kwargs):
        return wx.Colour(*args, **kwargs)


    def set_tooltip(ctrl, tooltip):
        ctrl.SetToolTip(tooltip)
else:
    newID = wx.NewId


    def wxCursor(*args):
        return wx.StockCursor(*args)


    def changeLightness(c, n):
        """:type c: wx.Colour"""
        if not isinstance(c, wx.Colour):
            c = wxColour(c)
        r, g, b = c
        f = n / 100.0
        r = max(0, min(255, int(r * f)))
        g = max(0, min(255, int(g * f)))
        b = max(0, min(255, int(b * f)))
        return wx.Colour(r, g, b)


    def wxColour(*args, **kwargs):
        if args:
            if isinstance(args[0], wx.Colour):
                return args[0]
            if isinstance(args[0], (list, tuple)):
                return wx.Colour(*args[0], **kwargs)
            if isinstance(args[0], basestring):
                return wx.Colour(*colorstr_to_RGB(args[0]))
        return wx.Colour()


    def set_tooltip(ctrl, tooltip):
        ctrl.SetToolTipString(tooltip)


def colorstr(*arg):
    if isinstance(arg[0], basestring):
        return colorstr_to_RGB(arg[0])

    arg += arg.__class__([0, ] * (3 - len(arg)))
    if len(arg) == 3:
        return "#%02x%02x%02x" % tuple(arg)
    return "#%02x%02x%02x%02x" % tuple(arg[:4])


def color_gray(level):
    if isinstance(level, float):
        level = int(255 * level)
    level = min(255, max(0, level))
    return [level, level, level]


def colorstr_to_RGB(s):
    """
    Returns R, G, B colors (0-255) based on a color string (hex notation
    with a leading "#"). Later wxPython implementation can deal with this themselfs.
    E.g.:
        "#ffffff" returns 255, 255, 255
        "#0f0f0f" returns 15, 15, 15

    :param s: Colorstring in hex with preceeding "#"
    :return: R, G, B
    :rtype: tuple
    """
    if len(s) > 7:
        return colorstr_to_RGBA(s)
    if s[0] == "#":
        s = s[1:]
        return int(s[:2], 16), int(s[2:4], 16), int(s[4:6], 16)
    raise AttributeError("Not a color string: '%s'" % s)


def colorstr_to_RGBA(s):
    """
    Same as colorstr_to_RGB() but with addiditional Alpha.
    :return: R, G, B, A
    """
    if s[0] == "#":
        s = s[1:]
        return [int(s[:2], 16), int(s[2:4], 16), int(s[4:6], 16), int(s[6:8])]
    raise AttributeError("Not a color string: '%s'" % s)


def list_of(listlen, arg, fill=None):
    """
    Return a list with exactly n arguments based on arg.
    Missed elements are set to the value provided in <fill>.
    """
    if not isinstance(arg, (tuple, list)):
        arg = [arg, ]
    return arg + arg.__class__([fill, ] * (listlen - len(arg)))


PopupEvent, EVT_POPUP_SELECTION = newevent.NewEvent()
